(*
 Oukseh Lee, The LET Project, KAIST

 Copyright(c) 2000 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)


structure Labm = struct
  structure Label = struct type t = StringInfo.labelinfo val compare = compare end
  structure M = OrderedListMapFn(Label)
  type key = Label.t
  type 'a t = 'a M.t

  exception InvalidIter2

  fun union s1 s2 = M.fold M.add s1 s2
  val length = List.length
  fun is_empty m = m = M.empty
  fun iter2 f [] [] = ()
    | iter2 f ((l1,t1)::o1) ((l2,t2)::o2) = (f l1 t1 l2 t2; iter2 f o1 o2)
    | iter2 f _ _ = raise InvalidIter2
  fun split m1 m2 = let
    fun f m k t (c,r) = if M.mem k m then (M.add k t c,r) else (c,M.add k t r)
    val (c1,rm1) = M.fold (f m2) m1 (M.empty,M.empty)
    val (c2,rm2) = M.fold (f m1) m2 (M.empty,M.empty)
   in
    ((c1,c2),(rm1,rm2))
   end   
  fun to_list m = m
  fun of_list l = List.fold_left (fn m (k,v) => M.add k v m) M.empty l
end 

structure Ty = struct
  open StringInfo
  open String_ast.Ast
  open Format
  
  type attr = IR | IRCS
  type ty = 
         Link of ty ref
       | Fun of ty * ty
       | Data of ty list * tyname
       | Tuple of ty list
       | Var of tyvar
       | Rec of ty Labm.t * (row ref) option
   and tyvar =
         Meta of int * int (* id / level *)
       | Ol of int * attr (* 0 is bound *)
       | Name of int * string 
       | Bound of int
   and row =
         Rvar of int
       | Row of ty Labm.t * (row ref) option

   and tyname = tyidinfo * int * int
                (* name / identifier / arity *)
   and cons = (conidinfo * ty option) list

   
  type scheme = int * ty

  structure Tids = struct
    open Set
    structure M = Make(struct type t = tyidinfo val compare = compare end)
    fun singleton (x,_) = M.singleton x
    fun add (x,_) s = M.add x s
  end

  structure Tv = struct
    type t = ty
    val count = ref 0
    fun new() = (count++; Link(ref(Var(Meta(!count,!count)))))

    structure Ol = struct
      val m = ref ([] : ty ref list)
      fun new attr = (count++; let val v = ref(Var(Ol(!count,attr))) 
                               in  m := v::(!m);
                                   Link v
                               end)
      fun resolve() = (List.iter (fn rty => case !rty of 
                                    Var(Ol _) => rty := Data([],("int",0,0))
                                  | _ => ()) (!m);
                       m := [])
    end
    fun new_named s = (count++; Var(Name((!count),s))) 
    fun current() = !count
    fun bound i = Var(Bound i)
    fun make_tvname n = 
      let fun f n = if n<26 then String.make 1 (Char.chr(97+n))
                            else (f (n/26))^(f (n mod 26))
      in  "'"^f(n)
      end

    fun pp_att IR = "n"
      | pp_att _  = "c"
    and pp x = case x of
          Meta (x,y) => if x=y then printf "_%d" x else printf "_%d(%d)" x y
        | Ol(i,att) => (printf "_%s" (pp_att att); printf "%d" i)
        | Bound i => printf "%s" (make_tvname i)
        | Name(_,s) => printf "%s" s 
  end

  structure Tn = struct
    type t = tyname
    val compare = compare
    val count = ref 0
    fun new s k = (s, count++; !count, k)
    fun current() = !count
    fun copy (s,_,k) = new s k
    fun arity (_,_,n) = n
    fun name (n,_,_) = n
    fun pp (s,i,x) = printf "%s" s 
  end

  structure Tns = ListSetFn(Tn)

  fun list_init n f = let
       fun g m = if n=m then [] else (f m)::(g(m+1))
      in g 0 end
  fun init_tuple n f = Tuple(list_init n f)

  structure Row = struct
    val count = ref 0
    fun new() = (count++; Some(ref(Rvar(!count))))
    fun closerow m = Rec(m,None)
    fun openrow m = Rec(m,new())
    val openrow_from_olist = openrow
    fun flatten x = let
      fun f r = case !r of
            Rvar _ => (Labm.M.empty, Some r)
          | Row(m,None) => (m,None)
          | Row(m,Some r0) => let val (m1,r1) = f r0 in (Labm.union m m1,r1) end
     in
      case x of (m,None) => x
              | (m,Some r) => (case (!r) of Rvar _ => x
                                          | _ => f(ref(Row x)))
     end

    (* Convert a record to a tuple 
       tuplize: convert to a tuple if possible (an open record is not converted)
       enforce_tuple: convert to a tuple (an open record is closed) *)
    exception CannotTuplize
    fun labm2tuple m = init_tuple (Labm.length m) (fn i=>Labm.M.find (string_of_int i) m)
    fun close r = case flatten r of (m,Some r) => (r := Row([],None); (m,None))
                                  | x => x 
    fun tuplize r = case flatten r of 
          (m,None) => ((labm2tuple m) handle _ => raise CannotTuplize)
        | _ => raise CannotTuplize
    fun enforce_tuple r = tuplize(close r)
  end

  val dummy = Tv.new
(*
  val static_dummy = Link(ref(Var(Meta(-1,-1))))
*)
  
  fun gen_data x name = Data(x,name)
  fun new_data x name = Data(x,Tn.new name (List.length x))

  fun gen_predef s= gen_data [] (s,0,0)
  val type_int =    gen_predef "int"
  val type_real =   gen_predef "real"
  val type_string = gen_predef "string"
  val type_bool =   gen_predef "bool"
  val type_unit =   gen_predef "unit"
  val type_char =   gen_predef "char"
  val type_exn =    gen_predef "exn"

  fun gen_predef t s = gen_data t (s,0,1)
  fun gen_ref t =   gen_predef [t] "ref"

  fun gen_tuple l = Tuple l
  fun gen_pair t1 t2 = gen_tuple [t1,t2]
  fun gen_list t =  gen_predef [t] "list"
  fun gen_array t = gen_predef [t] "array"
  fun gen_fun t1 t2 = Fun(t1,t2)

  val fold = List.fold_left

  fun pp_list f g nil = ()
    | pp_list f g (h::nil) = f h
    | pp_list f g (h::t) = (f h; g h; pp_list f g t)

  fun pp ty = let
    fun f t = case t of
        Link t' => pp (!t') (* (printf "|"; pp (!t'); printf "|") *)
      | Var v => Tv.pp v
      | Fun(t1,t2) => (
          (case t1 of Fun _ => (printf "@[("; f t1; printf ")@]")
                    | _ => f t1);
          printf "@ ->@ "; f t2)
      | Data(args,tn) => let
          fun complex t = case t of 
              Link l => complex (!l)
            | Tuple _ => true
            | Fun _ => true
            | _ => false
          val par = case args of 
                [] => false 
              | [arg] => complex arg
              | _ => true
         in 
          if par then printf "(";
          pp_list f (fn _ => printf ",@ ") args;
          if par then printf ")";
          if List.length args > 0 then printf "@ ";
          Tn.pp tn
         end
      | Rec x => let val (m,r) = Row.flatten x in
          printf "@[<1>{"; 
          pp_list (fn(k,t) => printf "%s: " k; pp t) (fn _ => printf ",@ ") (Labm.to_list m);
          case r of Some _ => printf "..." | _ => ();
          printf "}@]"
         end
      | Tuple tuple => pp_list f (fn _ => printf "@ * ") tuple
   in printf "@["; f ty; printf "@]" end

  fun is_expansive vbl = 
    let fun f e = case e of
            UnitExp _ | IntExp _ | RealExp _ | StringExp _ | CharExp _
          | VarExp _ | FnExp _ | ConExp _ => false
          | RecordExp(ll,_) => List.exists (fn (_,e,_)=> f e) ll
          | TupleExp(el,_)
          | ListExp(el,_) => List.exists f el
          | AppExp(e',e,_) => g e' || f e
          | ConstraintExp(e,_,_) | SeqExp([e],_) => f e
          | _ => true
        and g e = case e of
            ConExp _ => false
          | ConstraintExp(e,_,_) => g e
          | _ => true
    in  List.exists (fn (_,e,_) => f e) vbl 
    end

  fun unlink x = case x of
        (Link t) => (case !t of Var(Meta _) | Var(Ol _) => x | x => unlink x)
      | _ => x

 
 open Hashtbl
 structure Intmap = Make(struct type t=int fun equal x y = x=y val hash = Hashtbl.hash end)

 fun clos c t = let
(*   val _ = (printf "clos called (%d, " c; pp t; printf ").@.") *)
   val m = (Intmap.create 30): int Intmap.t
   val count = ref 0
   fun lookup_bound t = (Tv.bound(Intmap.find m t)) 
         handle Not_found => let val i = !count 
                             in count++; Intmap.add m t i; Tv.bound i end 
   fun f t = case t of
         Link t2 => (case !t2 of
           Var(Meta(x,y)) => if y<=c then t else lookup_bound x
         | Var(Ol _) => t
         | t => f t)
       | Fun(t1,t2) => Fun(f t1,f t2)
       | Data(ts,n) => Data(fs ts,n)
       | Tuple a => Tuple(fs a)
       | Rec r => Rec(case Row.close(Row.flatten r) of
           (m,None) => (Labm.M.map f m,None)
         | _ => raise Debug.CompilerBug "Ty.clos.Row")
       | Var(Bound x) => lookup_bound (-x)
       | Var(Name(x,_)) => if x<=c then t else lookup_bound x
       | Var _ => raise Debug.CompilerBug "Ty.bound.Var"
   and fs ts = List.map f ts
   val ft = f t
  in (!count,ft) end

 fun copy tvl t = let
   val n = List.length tvl
   val m = Array.of_list tvl
   fun f t = case t of
         Link t2 => (case !t2 of
           Var(Meta _)
         | Var(Name _) => t
         | (t as Var(Ol(0,_))) => f t
         | Var(Ol _) => t
         | t => f t)
       | Fun(t1,t2) => Fun(f t1,f t2)
       | Data(ts,n) => Data(fs ts,n)
       | Tuple a => Tuple(fs a)
       | Rec r => Rec(case r of
           (m,None) => (Labm.M.map f m,None)
         | _ => (case Row.flatten r of (m,x) => (Labm.M.map f m,x)))
       | Var(Bound i) => if i>=0 andalso i<n then m.[i] 
                         else raise Debug.CompilerBug "Ty.instance.Var(Bound _)"
       | Var(Ol(0,a)) => (case m.[0] of 
                            Link t2 => ((case !t2 of
                              Var(Meta _) => m.[0] <- Tv.Ol.new a 
                            | _ => ()); m.[0])  
                          | _ => raise Debug.CompilerBug "Ty.instance.Var(Ol _)")
       | _ => t
   and fs ts = List.map f ts
  in 
     (f t) handle _ => raise Debug.CompilerBug "Ty.copy"
  end

 fun instance (n,t) = copy (list_init n (fn _ => Tv.new())) t
    
 fun in_type (_,t) = t
 fun in_scheme t = (0,t)

 fun used_tyids t = let
   fun f (ConstTy(tl,([],id,_),_)) = Tids.M.union (fs tl) (Tids.singleton id)
     | f (FunTy(t1,t2,_)) = Tids.M.union (f t1) (f t2)
     | f (RecordTy(l,_)) = fr l
     | f (TupleTy(l,_)) = fs l
     | f _ = Tids.M.empty
   and fs l = List.fold_right (fn t => Tids.M.union (f t)) l Tids.M.empty
   and fr l = List.fold_right (fn (_,t,_) => Tids.M.union (f t)) l Tids.M.empty
  in f t end

 structure Tvs = ListSetFn(struct type t = tyid 
                                  val compare = fn (x,_) (y,_) => compare x y end)
 fun used_tyvars t = let
   fun f (ConstTy(l,_,_)) = fs l
     | f (TupleTy(l,_)) = fs l
     | f (FunTy(t1,t2,_)) = Tvs.union (f t1) (f t2)
     | f (RecordTy(l,_)) = fr l
     | f (VarTy(id,_)) = Tvs.singleton id
   and fs l = List.fold_right (fn t => Tvs.union (f t)) l Tvs.empty
   and fr l = List.fold_right (fn (_,t,_) => Tvs.union (f t)) l Tvs.empty
  in f t end

 exception Not_equal 

 (* assume that input types satisfies the constraints after type inference *) 
 fun equal t1 t2 = 
   let fun f t1 t2 = case (t1,t2) of
         (Link t1',Link t2') => f (!t1') (!t2')
       | (Link t1',_) => f (!t1') t2
       | (_,Link t2') => f t1 (!t2')
       | (Fun(t11,t12), Fun(t21,t22)) => (f t11 t21; f t12 t22)
       | (Tuple l1, Tuple l2) => List.iter2 f l1 l2
       | (Tuple _,Rec r) => f t1 (Row.enforce_tuple r)
       | (Rec r,Tuple _) => f (Row.enforce_tuple r) t2
       | (Rec r1,Rec r2) => (case (Row.flatten r1,Row.flatten r2) of
                               ((m1,None),(m2,None)) => 
                                  Labm.iter2 (fn l1 t1 l2 t2 => 
                                                if l1=l2 then f t1 t2
                                                         else raise Not_equal) 
                                              m1 m2
                             | _ => raise Debug.CompilerBug "Ty.equal.Rec")
       | (Data(l1,n1), Data(l2,n2)) => if n1=n2 then List.iter2 f l1 l2 
                                                else raise Not_equal
       | (x,y) => if x<>y then raise Not_equal
   in  (f t1 t2; true) handle (x as Debug.CompilerBug _) => raise x | _ => false 
   end

  structure CE = struct
    type cons = (conidinfo * ty option) list
    structure M = MapFn(Tn)
    type t = cons M.t
    val listcons = let val a = Tv.bound 0
     in [("nil",None),("::",Some(gen_pair a (gen_list a)))]
     end
    val initce = M.add ("list",0,1) listcons M.empty
    val m' = ref (initce:t)
    val m = ref (M.empty:t)
    fun init() = (m' := initce; m := M.empty)
    fun find tn = (M.find tn (!m)) handle Not_found => (
                  (M.find tn (!m')) handle Not_found => [])
    fun add tn [] = ()
      | add tn cons = m := M.add tn cons (!m)
    fun add_ce ce = M.iter (fn k v => m' := M.add k v (!m')) ce
    fun add_ce' ce = M.iter (fn k v => m := M.add k v (!m)) ce

    fun print_all() = let
      fun pp_cons(a,None) = printf "%s@ " a 
        | pp_cons(a,Some t) = (printf "%s of "a; pp t; printf "@ ") 
      fun f k v = (Tn.pp k; printf " = @["; List.iter pp_cons v; printf "@]@.")
     in M.iter f (!m) end
  end

  structure TS = struct
    type t = TsName of tyname (* data type *)
           | TsFun of scheme

    exception Apply_error

    fun arity ts = case ts of 
          TsName tn => Tn.arity tn
        | TsFun (n,_) => n

    fun normalize x = 
      let fun f i [] = true
            | f i (h::t) = if unlink h = Tv.bound i then f (i+1) t else false
      in  case x of
            TsFun(n,tau) => (case unlink tau of
              Data(tl,tn) => if Tn.arity tn = n andalso
                                List.length tl = n
                                andalso f 0 tl then TsName tn else x
            | _ => x)
          | _ => x
      end
   
    fun gen_data (id,_) n = TsName(Tn.new id n)
    fun gen_fun n tau = normalize(TsFun(n,tau))
    fun gen_name t = TsName t
    fun gen_var (id,_) n = TsName(Tn.new id n)

    fun tynames ts = case ts of
        TsName tn => Tns.singleton tn 
      | _ => Tns.empty

    val pp_ty = pp
    val pp = 
      let fun f ts = case ts of
            TsName x => (printf "n:"; Tn.pp x)
          | TsFun(n,t) => (printf "f(%d," n; pp t; printf ")")
      in  f end
    
    fun apply tl ts = 
         if arity ts = List.length tl then 
          case ts of
            TsName tn => Data(tl,tn)
          | TsFun (n,tau) => copy tl tau
         else raise Apply_error

  end
end
