(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  en Automatique.  All rights reserved.  This file is distributed    *)
(*  under the terms of the Q Public License version 1.0.               *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ntoploop.ml,v 1.1.1.1 2001/03/14 09:43:36 cookcu Exp $ *)

(* The interactive toplevel loop *)

open Lexing
open Format
open Config
open Misc
open Parsetree
open Types
open Typedtree
open Nprintval
open N2caml

type directive_fun =
   | Directive_none of (unit -> unit)
   | Directive_string of (string -> unit)
   | Directive_int of (int -> unit)
   | Directive_ident of (Longident.t -> unit)
   | Directive_bool of (bool -> unit)

(* Hooks for parsing functions *)

let print_location = Location.print
let print_warning = Location.print_warning
let input_name = Location.input_name

(* Load in-core and execute a lambda term *)

type evaluation_outcome = Result of Obj.t | Exception of exn

let load_lambda ppf lam =
  if !Clflags.dump_rawlambda then fprintf ppf "%a@." Printlambda.lambda lam;
  let slam = Simplif.simplify_lambda lam in
  if !Clflags.dump_lambda then fprintf ppf "%a@." Printlambda.lambda slam;
  let (init_code, fun_code) = Bytegen.compile_phrase slam in
  if !Clflags.dump_instr then
    fprintf ppf "%a%a@."
    Printinstr.instrlist init_code
    Printinstr.instrlist fun_code;
  let (code, code_size, reloc) = Emitcode.to_memory init_code fun_code in
  let can_free = (fun_code = []) in
  let initial_symtable = Symtable.current_state() in
  Symtable.patch_object code reloc;
  Symtable.update_global_table();
  try
    let retval = (Meta.reify_bytecode code code_size) () in
    if can_free then Meta.static_free code;
    Result retval
  with x ->
    if can_free then Meta.static_free code;
    Symtable.restore_state initial_symtable;
    Exception x

(* Print the outcome of an evaluation *)

let pr_item env ppf = function
  | Tsig_value(id, decl) :: rem ->
      begin match decl.val_kind with
      | Val_prim _ ->
          fprintf ppf "@[<2>%a@]"
          (Nprinttyp.value_description id) decl
      | _ ->
          fprintf ppf "@[<2>%a =@ %a@]"
          (Nprinttyp.value_description id) decl
          (print_value env (Symtable.get_global_value id))
          decl.val_type
      end;
      rem
  | Tsig_type(id, decl) :: rem ->
     if (Ident.name id).[0] = '@' && ((Ident.name id) <> "@float") then rem
     else begin
      fprintf ppf "@[%a@]"
      (Nprinttyp.type_declaration id) decl;
      rem
     end
  | Tsig_exception(id, decl) :: rem ->
      fprintf ppf "@[%a@]"
      (Nprinttyp.exception_declaration id) decl;
      rem
  | Tsig_module(id, mty) :: rem ->
     if (Ident.name id).[0] = '@' then rem
     else (
      fprintf ppf "@[<2>structure %a :@ %a@]"
      Nprinttyp.ident id
      Nprinttyp.modtype mty;
      rem)
  | Tsig_modtype(id, decl) :: rem ->
      fprintf ppf "@[%a@]" 
      (Nprinttyp.modtype_declaration id) decl;
      rem
  | Tsig_class(id, decl) :: cltydecl :: tydecl1 :: tydecl2 :: rem ->
      fprintf ppf "@[%a@]"
      (Nprinttyp.class_declaration id) decl;
      rem
  | Tsig_cltype(id, decl) :: tydecl1 :: tydecl2 :: rem ->
      fprintf ppf "@[%a@]" 
      (Nprinttyp.cltype_declaration id) decl;
      rem
  | _ -> []

let rec print_items env ppf = function
  | [] -> ()
  | items ->
     match pr_item env ppf items with
     | [] -> ()
     | items -> fprintf ppf "@ %a" (print_items env) items;;

(* The current typing environment for the toplevel *)

let toplevel_env = ref Env.empty

(* Print an exception produced by an evaluation *)

let print_exception_outcome ppf = function
  | Sys.Break ->
      fprintf ppf "Interrupted.@."
  | Out_of_memory ->
      Gc.full_major();
      fprintf ppf "Out of memory during evaluation.@."
  | Stack_overflow ->
      fprintf ppf "Stack overflow during evaluation (looping recursion?).@."
  | exn ->
      fprintf ppf "@[Uncaught exception:@ %a.@]@."
              (print_value !toplevel_env (Obj.repr exn)) Predef.type_exn

(* The table of toplevel directives. 
   Filled by functions from module topdirs. *)

let directive_table = (Hashtbl.create 13 : (string, directive_fun) Hashtbl.t)

(* Execute a toplevel phrase *)

let execute_phrase print_outcome ppf phr =
  match phr with
  | Ptop_def sstr ->
      let (str, sg, newenv) = Ntypemod.type_structure !toplevel_env sstr in
      let lam = Translmod.transl_toplevel_definition str in
      let res = load_lambda ppf lam in
      begin match res with
      | Result v ->
          if print_outcome then begin
            match str with
            | [Tstr_eval exp] ->
                fprintf ppf "@[- : %a@ =@ %a@]@."
                Nprinttyp.type_scheme exp.exp_type
                (print_value newenv v) exp.exp_type
            | [] -> ()
            | _ ->
                fprintf ppf "@[<v>%a@]@."
                (print_items newenv) sg
          end;
          toplevel_env := newenv;
          true
      | Exception exn ->
          print_exception_outcome ppf exn;
          false
      end
  | Ptop_dir(dir_name, dir_arg) ->
      try
        match (Hashtbl.find directive_table dir_name, dir_arg) with
        | (Directive_none f, Pdir_none) -> f (); true
        | (Directive_string f, Pdir_string s) -> f s; true
        | (Directive_int f, Pdir_int n) -> f n; true
        | (Directive_ident f, Pdir_ident lid) -> f lid; true
        | (Directive_bool f, Pdir_bool b) -> f b; true
        | (_, _) ->
            fprintf ppf "Wrong type of argument for directive `%s'.@." dir_name;
            false
      with Not_found ->
        fprintf ppf "Unknown directive `%s'.@." dir_name;
        false

(* Temporary assignment to a reference *)

let protect r newval body =
  let oldval = !r in
  try
    r := newval; 
    let res = body() in
    r := oldval;
    res
  with x ->
    r := oldval;
    raise x

(* Read and execute commands from a file *)

let use_print_results = ref true


(* Reading function for interactive use *)

let first_line = ref true
let got_eof = ref false;;

let refill_lexbuf buffer len =
  if !got_eof then (got_eof := false; 0) else begin
    output_string stdout (if !first_line then "# " else "  "); flush stdout;
    first_line := false;
    let i = ref 0 in
    try
      while true do
        if !i >= len then raise Exit;
        let c = input_char stdin in
        buffer.[!i] <- c;
        incr i;
        if c = '\n' then raise Exit;
      done;
      !i
    with
    | End_of_file ->
        Location.echo_eof ();
        if !i > 0 then (got_eof := true; !i) else 0
    | Exit -> !i
  end

(* Discard everything already in a lexer buffer *)

let empty_lexbuf lb =
  let l = String.length lb.lex_buffer in
  lb.lex_abs_pos <- (-l);
  lb.lex_curr_pos <- l

